/*
***************************************************************************************
 Subwoofer Test Generator - 10 Hz to 90 Hz sine wave in 0.1 Hz or 1.0 Hz resolution
 Version 1.1
 Copyright 2021, 2022
 Jed Margolin

 Permission is granted to individuals to make copies for their personal use.
 Any commercial use requires my permission

 Board: Woof Rev 3  Low Frequency Generator
 Uses LCD Display in 4-bit mode)
 Uses factory DCO values; no crystal

 Use 0 - Register Optimizations

 MSP430G2402 or greater

 Built with Code Composer Studio v10

 INC - Press and release to increase the frequency
     - Press and hold for about 2 seconds to keep increasing the frequency; release to stop

 DEC - Press and release to decrement the frequency
     - Press and hold for about 2 seconds to keep decreasing the frequency; release to stop

 MODE - Press and release to toggle the frequency step between 0.1 Hz and 1.0 Hz
      - Press and hold for about 2 seconds (then release) to enter Sweep mode at 0.1 Hz increments
      - When in Sweep mode Press and hold for about 2 seconds (then release) to exit Sweep mode

 Port 1:
   D0  input         SW1 (INC)
   D1  input         SW2 (DEC)
   D2  input         SW3 (MODE)

   D3  output        /LDAC  Load DAC (MPC4911)
   D4  output        /CS    Chip Select
   D5  SPI output    SCLK
   D6  SPI output    SDOUT
   D7  SPI input     SDIN  (Not used by MCP4911)

 Port 2:
   D0  output    LCD D4
   D1  output    LCD D5
   D2  output    LCD D6
   D3  output    LCD D7
   D4  output    LCD /E (Enable)
   D5  output    LCD RS (Register Select)

   D6  input     XIN  (no crystal)
   D7  input     XOUT (no crystal
***************************************************************************************
*/

#include <msp430.h>
#include "freq64table.h"
#include "main.h"
//================================================
int main(void)
{
    unsigned int k;
    int freq_step;

    WDTCTL = WDTPW + WDTHOLD;   // Stop watchdog timer

// No crystal. Use the factory default values for the DCO
    DCOCTL = CALDCO_16MHZ;      // factory calibration vqlues
    BCSCTL1 = CALBC1_16MHZ;

    BCSCTL2 = 0x00;              // DIVM0 | DIVS0 |   // MCLK and SMCLK dividers

    P1SEL  &= ~(BIT0 | BIT1 | BIT2);     // In   SW1 (Increment); SW2 (Decrement); SW2 (Mode)
    P1SEL2 &= ~(BIT0 | BIT1 | BIT2);
    P1REN |= (BIT0 | BIT1| BIT2);        // pullup/pulldown enabled
    P1OUT |= (BIT0 | BIT1 | BIT2);       // pullup
    P1DIR &= ~(BIT0 | BIT1 | BIT2);      // In

    P1SEL  &= ~(BIT3);      // Out - DAC CS
    P1SEL2 &= ~(BIT3);
    P1OUT |= BIT3;
    P1DIR |= BIT3;          // Out

    P1SEL  &= ~(BIT4);      // Out - DAC Latch
    P1SEL2 &= ~(BIT4);
    P1OUT |= ~(BIT4);
    P1DIR |= BIT4;          //

// Port 2 for LCD
    P2SEL  &= ~(0x3F);      // Out
    P2SEL2 &= ~(0x3F);
    P2DIR |= 0x3F;          // Out
    P2OUT &= 0x3F;          // LCD D4-7 = 0; LCD E = 0; LCD RS = 0

// Port 2 - We are not using a crystal so make D6 and D7 inputs (pulled up)
    P2SEL  &= ~(BIT6 | BIT7);         // In   Xin, Xout
    P2SEL2 &= ~(BIT6 | BIT7);
    P2DIR &= ~(BIT6 | BIT7);
    P2REN |= (BIT6 | BIT7);          // pullup/pulldown enabled
    P2OUT |= (BIT6 | BIT7);          // pullup

    sweep = 0;
    freq_step = 10;            // frequency increment: 10 = 1Hz; 1 = 0.1Hz

// Set up SPI interface
    spi_init();

// Set up Interrupts
    CCTL0 = CCIE;                      // CCR0 interrupt enabled
    table_index = 300;                 // 40.0 Hz
    CCR0 = table_index;
    TACTL = TASSEL_2 + ID_0 + MC_1;    // 0x210  SMCLK, /1, Upmode
    TACTL |= TACLR;                    // clear timer Resets TAR, clock divider, count direction

    __bis_SR_register(GIE);           // enable interrupts

// Initialize LCD
    lcd_init();
//==================================
    while (1)           // main loop
    {
       k = table_index + 100;   // Display the frequency
       lcd_wr_bcd(0xC0,k);      // Takes 4.7 ms

// Debounce and process the switches
       do_switch(P1IN, BIT0, &switch1);          // PortIn, Bit mask, switch structure
       do_switch(P1IN, BIT1, &switch2);          // PortIn, Bit mask, switch structure
       do_switch(P1IN, BIT2, &switch3);          // PortIn, Bit mask, switch structure

// We now know if a switch has been pressed (and how long it has been pressed)
//    and if a switch has been released

// Operating modes

// SW1
       if(switch1.press == 1)
       {
          table_index = table_index + freq_step;            // increment the frequency
          if(table_index >800) {table_index = 0;}
          switch1.press = 0;         // reset the flag
          switch1.count = 0;
          switch1.countrun = 1;      // the counter is running
       }

       if(switch1.countrun && (switch1.count <200) )
        {
            switch1.count++;
        }

        if(switch1.countrun && (switch1.count >=200) )
        {
            switch1.count = 160;        // limit the count

            table_index = table_index + freq_step;
            if(table_index >800)
            {
                table_index = 0;
            }
        }

        if(switch1.release)         // stop and reset the counter
        {
            switch1.release = 0;
            switch1.count = 0;
            switch1.countrun = 0;
        }

// SW2
       if(switch2.press)
       {
          table_index = table_index - freq_step;        // decrement the frequency
          if(table_index < 0) {table_index = 800;}

          switch2.count = 0;
          switch2.press = 0;
          switch2.countrun = 1;
       }

       if( switch2.countrun && (switch2.count <200) )
       {
           switch2.count++;
       }

       if( switch2.countrun && (switch2.count >=200) )
       {
           switch2.count = 160;

           table_index = table_index - freq_step;
           if(table_index <0)
           {
               table_index = 800;
           }
       }

       if(switch2.release)
       {
           switch2.release = 0;
           switch2.count = 0;
           switch2.countrun = 0;
       }

// SW3
       if(switch3.press)
       {
          switch3.count = 0;
          switch3.press = 0;
          switch3.countrun = 1;

          // change the frequency step here
          if(freq_step == 1)  {freq_step = 10;}   // 1 = 0.1Hz; 10 = 1Hz
          else
          {
             if(freq_step == 10) {freq_step = 1;}
          }
       }

       if(switch3.countrun && (switch3.count <150) )   // was 200
       {
           switch3.count++;
       }

      if(switch3.countrun && (switch3.count >=150) )   // was 200
      {
          sweep ^= 0x01;            // switch between sweep and manual control
          switch3.countrun = 0;     // turn it off
      }

      if(switch3.release)   // stop and reset the counter
      {
         switch3.release = 0;
         switch3.count = 0;
         switch3.countrun = 0;
      }

// Display Frequency Increment
      if(sweep)
      {
         domsg(&msg3);        // "Sweep "
      }
      else
      {
         if(freq_step == 1)
         {
            domsg(&msg4);        // "0.1 Hz"
         }

         if(freq_step == 10)
         {
            domsg(&msg5);        // "1.0 Hz"
         }

        if( (freq_step != 10) && (freq_step != 1))
        {
           domsg(&msg6);        // "oops! "
        }
      } // end else

    }  // end while main loop

} // end main
//==========================================================
// This is where we step through the sine table and send it to the DAC
// Timer A0 interrupt service routine
#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector=TIMER0_A0_VECTOR
__interrupt void Timer_A (void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(TIMER0_A0_VECTOR))) Timer_A (void)
#else
#error Compiler not supported!
#endif
{
   static int k=0;
   static int n=0;
   static int pindex=0;
   int txword;

   txword = sine[k];
   txword |= 0x3000;        // for MCP4921  d15 0=active, d14 0 = Vref unbuffered, d13 1= gain of 1, d12 1=active

   spi_sendword(txword);

   k = (k + 1) & 0x3F;      // 64 samples/period

// test period table
   if(sweep)
   {
      if(n++ == 80)
      {
         n = 0;
         pindex++;
         if(pindex>800) {pindex= 0;}
         CCR0 = period[pindex];
         table_index = pindex;          // for displaying the frequency
      }
   }
   else         // Period comes from mainline selected by switches
   {
      pindex = table_index;
      CCR0 = period[pindex];
   }
}
//=================================================================
void wait(unsigned int time)        // 500 uS
{
    volatile unsigned int i,j;

    for(i=time; i>0; i--)
    {
        for(j=382; j>0; j--);
        {
        }
    }

 return;
}
//==========================================================
void wait4(unsigned int time)        // 500 uS
{
    volatile unsigned int i,j;

    for(i=time; i>0; i--)
    {
        for(j=20; j>0; j--);
        {
        }
    }

 return;
}
//=================================================================
void spi_init(void)
{

    USICTL1 |= USISWRST | USICKPH;      // Reset, invert clock phase

    USICTL0 |= USIPE7 | USIPE6 | USIPE5 | USIMST | USIOE;    // Set USICTL_0  MSB first, Master

    USICKCTL = USIDIV_0 + USISSEL_2;    // Set USICKCTL  SMCLK
    USICNT = 8;                         // 8 bit data length

// SPI CLK
    P1SEL  |= BIT5;         //  1:
    P1SEL2 &= ~(BIT5);      //  0;

// SDO
    P1SEL  |= BIT6;         // 1:
    P1SEL2 &= ~(BIT6);      // 0:

// SDI
    P1SEL  |= BIT7;         // 1:n
    P1SEL2 &= ~(BIT7);      // 0:

    USICTL0 &= ~USISWRST;               // USI released for operation

    USICNT = 8;                         // Init-load counter

}

void spi_sendword(unsigned int txword)
{
    P1OUT &= ~BIT4; // DAC CS

    while (!(USIIFG & USICTL1));            // Counter clear?
    USISR = (txword >> 8);
    USICNT = 8;     // re-load counter

    while (!(USIIFG & USICTL1));            // Counter clear?
    USISR = txword & 0xFF;                  // low byte;
    USICNT = 8;     // re-load counter

    while (!(USIIFG & USICTL1));            // Counter clear?

    P1OUT |= BIT4;  // DAC CS

    P1OUT &= ~(BIT3);    // DAC Latch
    P1OUT |= BIT3;    // DAC Latch
}

void spi_sendbyte(unsigned char txbyte)
{
    P1OUT &= ~BIT4;

    while (!(USIIFG & USICTL1));            // Counter clear?
    USISR = txbyte;
    USICNT = 8;     // re-load counter

    while (!(USIIFG & USICTL1));            // Counter clear?

    P1OUT |= BIT4;
}
// ====================================
void lcd_init(void)
{
    lcd_reset();

    domsg(&msg1);        // "Subwoofer Tester"
    domsg(&msg7);        // "Version 1.1"
    wait(8000);
    clear_lcd(0xC0);      // Second line of LCD
    domsg(&msg8);        // "Jed Margolin"
    wait(8000);
    clear_lcd(0xC0);      // Second line of LCD
}
//=========================
void lcd_reset(void)
{
/*
P2.0 - LCD D4
P2.1 - LCD D5
P2.2 - LCD D6
P2.3 - LCD D7

P2.4 - LCD E (Enable High, Write Data, Enable Low)
P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
*/

// Wait at least 16 ms after power-up
    wait10ms();    // wait 10 ms
    wait10ms();    // wait 10 ms

    P2OUT &= (~BIT5);    // RS = 0

    lcd_wr4(0x03);      // write nybble and wait 200us
    wait10ms();

    lcd_wr4(0x03);      // write nybble and wait 200us
    lcd_wr4(0x03);      // write nybble and wait 200us
    lcd_wr4(0x02);      // write nybble and wait 200us
    lcd_wr_instr(0x28); //  2 lines, 5x7 dots
    lcd_wr_instr(0x08); //  display off
    lcd_wr_instr(0x01); //  display on

    wait10ms();

//    lcd_wr_instr(0x0F); //  display on, cursor on, blink on
    lcd_wr_instr(0x0c);  //  display on, cursor oFF, blink oFF

    lcd_wr_instr(0x06);  //  entry mode set: cursor increment, no shift
}
//==================================================================
// Clear one line
void clear_lcd(unsigned char pos)
{
    unsigned int n;

    lcd_wr_instr(pos);

    for(n=16; n>0; --n)
    { lcd_wr_data(' ');}
}
//===============================================================
// Do LCD Message
//void domsg(struct messages *sptr)
void domsg(const struct messages *sptr)
{
    unsigned char m,n;
    unsigned char data;

    m = sptr->pos;  // position
    lcd_wr_instr(m);

    m = sptr->numchars;  // numbers of chars

    for(n=0; n<m; n++)
    {
        data = sptr->msg[n];
        lcd_wr_data(data);
    }
}
//=================================================================
// wait 10 ms
void wait10ms(void)
{
   wait(41);
}
//======================
void lcd_wr4(unsigned int data)      // write nybble and wait 200us
{

//    P2.4 - LCD E (Enable High, Write Data, Enable Low)
//    P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)

    P2OUT |= BIT4;     // Enable High

    data &= 0x0F;
    P2OUT &= 0xF0;
    P2OUT |= data;     // write nybble

    P2OUT &= (~BIT4);  //  /E low

    wait4(20);         // 200 us
}
//======================
// LCD - Write 8-bit instruction
// P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
void lcd_wr_instr(unsigned int data)
{
   unsigned int dsave;

   P2OUT &= (~BIT5);    // RS = 0

   dsave = data;
   dsave  = dsave >>4;       // high nybble
   lcd_wr4(dsave);  // write nybble and wait 200 us

   data &=0xF;      // low nybble4
   lcd_wr4(data);
}
//======================
// LCD - Write 8-bit data
// P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
void lcd_wr_data(unsigned int data)
{
    unsigned int dsave;

    P2OUT |= BIT5;    // RS = 1

    dsave = data;
    dsave = dsave >>4;       // high nybble
    lcd_wr4(dsave);  // write nybble and wait 200 us

    data &=0xF;      // low nybble4
    lcd_wr4(data);
}
//======================
void lcd_wr_bcd(unsigned int pos, unsigned int bin)
{

// First Convert Binary to ASCII
// Adapted from 8-bit microcontroller code using a DAA instruction

// unsigned int bin = 0x386; // 0x67 = 103d; 0x386 = 902d
   unsigned char bcd0 = 0;
   unsigned char bcd1 = 0;
   unsigned char bcd2 = 0;
   unsigned char char1, char2, char3;
   int i;
   unsigned char hcarry;

   for(i=0; i<12; i++)
   {
//---------- BCD0
      bcd0 = bcd0<<1;

      if(bin & 0x800)
      {bcd0 = bcd0 +1;}

      bin = bin<<1;

      if(bcd0 >9)
      {
         bcd0 = bcd0 - 10;
         hcarry = 1;
      }
      else {hcarry = 0; }

//---------- BCD1
      bcd1 = bcd1<<1;
      bcd1 = bcd1 + hcarry;

      if(bcd1 >9)
      {
          bcd1 = bcd1 - 10;
          hcarry = 1;
      }
      else {hcarry = 0;}

//---------- BCD2
      bcd2 = bcd2<<1;
      bcd2 = bcd2 + hcarry;

     if(bcd2 >9)
     {
         bcd2 = bcd2 - 10;
//       hcarry = 1;
     }
//        else {hcarry = 0;}

   }

// Convert to ascii
   char1 = bcd2 + 0x30;
   char2 = bcd1 + 0x30;
   char3 = bcd0 + 0x30;

// Display it
   lcd_wr_instr(pos);
   lcd_wr_data(char1);
   lcd_wr_data(char2);
   lcd_wr_data('.');
   lcd_wr_data(char3);
   lcd_wr_data(' ');
   lcd_wr_data('H');
   lcd_wr_data('z');

}
//=========================
void do_switch(unsigned char port, unsigned char bit, struct myswitch *sptr)    // PortIn, Bit, switch structure
{
//    unsigned char temp;

// Debounce and process the switch
// sptr points to the switch's data block
    sptr->debounce  = sptr->debounce <<1;    // shifts in a zero
    if(port & bit)        // a zero has already been shifted in
    {
       sptr->debounce++ ;      // so make LSB a one
    }

// Debounced state is either 0x00 or 0xFF
    if(sptr->debounce == 0x00)  {sptr->state = 0x00;}
    if(sptr->debounce == 0xFF)  {sptr->state = 0x01;}

// process the switch
    if(sptr->state != sptr->old)     // we have a change in switch state
    {
       if( (sptr->old == 0x01) && (sptr->state == 0x00) )  // we have a switch press
       {
          sptr->press = 0x01;
          sptr->count = 0;        // clear the switch counter
       }

       if( (sptr->old == 0x00) && (sptr->state== 0x01) )  // we have a switch release
       {
          sptr->release = 0x01;
          sptr->count = 0;        // clear the switch counter       }
       }

       sptr->old = sptr->state;     // the new state is now the old state
    }

// We now know if a switch has been pressed (and how long it has been pressed)
//    and if a switch has been released

    asm(" nop");    // for emulator
}
//===================================================

